/*
 *	(c) 1996,1997 - Mathijs Maassen, PE1NTP
 *
 *	April 1997
 *
 *	(Used TAB width: 4)
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#ifdef __linux__
#include <unistd.h>
#endif
#include <fcntl.h>
#include <string.h>
#include <ctype.h>
#include <setjmp.h>
#include <errno.h>
#include "fbb.h"
#include "reqlst.h"

#define REFUSE 0
#define ACCEPT 1
#define EXCEPT 2

#define ARG_STR 0
#define ARG_7PL 1
#define ARG_NUL 2

#define MAX_FLT 500
#define MAX_SUB 64

#define FLAG_7PLUS "\001"		/* msg contains 7+ data		*/

static int flagmat(char *, char *);
static int isempty(char *, char *);

static int num_filt;			/* number of filter entries	*/
static jmp_buf env;
static char saveflt_fmt[] = "REQLST/%s.flt";

int del_filt;
int save_filt;			/* save filter data	if !0	*/
int show_filt;			/* show filter data	if !0	*/
char flterr[80] = "";

static struct filter {
	int (*matchfunc)(char *, char *);
	int mode;
	char *field;
	char *str;
} **filter_entries;

static struct {
	char *cmdstr;
	int *val;
	int need_arg;
	int min_val;
} cmds[] = {
	{ "delete",  &del_filt,  0, 0  },
	{ "linelen", &line_len,  1, 46 },
	{ "maxlines",&max_lines, 1, 0  },
	{ "save",    &save_filt, 0, 0  },
	{ "show",    &show_filt, 0, 0  },
	{ NULL,      NULL,       0, 0  }
};

static struct {
	char dest[7];
	char bbs[7];
	char orig[7];
	char bid[13];
	char flags[1];
	char loc[1];
	char title[61];
} msg;

static struct {
	char *cmdstr;
	int (*func)(char *, char *);
	char *fieldp;
	int argtype;
} flt_cmd[] = {
	{ ">",    wildmat, msg.dest,  ARG_STR },
	{ "@",    wildmat, msg.bbs,   ARG_STR },
	{ "<",    wildmat, msg.orig,  ARG_STR },
	{ "s",    wildmat, msg.title, ARG_STR },
	{ "$",    wildmat, msg.bid,   ARG_STR },
	{ "7p",   flagmat, msg.flags, ARG_7PL },
	{ "loc",  isempty, msg.loc,   ARG_NUL },
	{ NULL,   NULL,    NULL,      0       }
};

static void do_init_err(char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	vsprintf(flterr, fmt, ap);
	va_end(ap);
	longjmp(env, 1);
}

static void *do_malloc(size_t size)
{
	void *p;

	if ((p = malloc(size)) == NULL)
		do_init_err("OOPS-ERROR: Out of memory (malloc)\n");
	return (p);
}

static void *do_realloc(void *ptr, size_t size)
{
	void *p;

	if ((p = realloc(ptr, size)) == NULL)
		do_init_err("OOPS-ERROR: Out of memory (realloc)\n");
	return (p);
}

static void do_cmd(char *str)
{
	int i, ret, iarg = 0;
	char cmd[10], arg[10];

	ret = sscanf(str, "%9s %9s", cmd, arg);

	for (i = 0; cmds[i].cmdstr; i++)
		if (!strcasecmp(cmd, cmds[i].cmdstr)) {
			if (cmds[i].need_arg) {
				if (ret != 2) {
					do_init_err("`%s' requires an argument.\n", cmd);
				} else {
					iarg = atoi(arg);
					if (iarg < cmds[i].min_val)
						do_init_err("Minimum value for argument of `%s' is %d.\n", cmd, cmds[i].min_val);
				}
			}
			*cmds[i].val = (cmds[i].need_arg) ? iarg : 1;
			return;
		}

	do_init_err("Invalid command: %s\n", cmd);
}

static int flagmat(char *p, char *val)
{
	return (*p & *val);
}

static int isempty(char *p, char *notused)
{
	return (*p == '\0');
}

static int add_entry(char *line)
{
	int mode, cmdcnt, sub_filt = 0, expect_cmd = 0, got_args = 0;
	char *p;

	if (line[1] != ' ')
		do_init_err("Syntax error in filter #%d. Space expected after action.\n", num_filt + 1);

	mode = ((*line == '-') ? REFUSE : ACCEPT);
	p = strtok(line, " \t\n");

	for (p = strtok(NULL, " \t\n"); p; p = strtok(NULL, " \t\n")) {
		got_args = 1;
		switch (expect_cmd) {
		case 0:	/* expect cmd */
			if (!strcmp(p, "!")) {
				if (!sub_filt)
					do_init_err("Syntax error in filter #%d. Didn't expect a '!'\n", num_filt + 1);
				mode = EXCEPT;
				continue;
			}
			for (cmdcnt = 0; flt_cmd[cmdcnt].cmdstr; cmdcnt++) {	/* find cmd */
				if (!strcasecmp(flt_cmd[cmdcnt].cmdstr, p)) {
					if (sub_filt == 0)
						filter_entries[num_filt] = do_malloc(MAX_SUB * sizeof(struct filter));
					filter_entries[num_filt][sub_filt].field = flt_cmd[cmdcnt].fieldp;
					filter_entries[num_filt][sub_filt].matchfunc = flt_cmd[cmdcnt].func;
					filter_entries[num_filt][sub_filt].mode = mode;
					if (flt_cmd[cmdcnt].argtype == ARG_STR) {
						expect_cmd = 1;
						break;
					} else if (flt_cmd[cmdcnt].argtype == ARG_7PL)
						filter_entries[num_filt][sub_filt].str = FLAG_7PLUS;
					else if (flt_cmd[cmdcnt].argtype == ARG_NUL)
						filter_entries[num_filt][sub_filt].str = "\0";
					if (++sub_filt == MAX_SUB)
						do_init_err("Too many fields on one line. (%d max.)\n", MAX_SUB);
					expect_cmd = 0;
					break;
				}
			}
			if (flt_cmd[cmdcnt].cmdstr == NULL)
				do_init_err("Invalid cmd or missing space in filter #%d, field #%d: %s\n", num_filt + 1, sub_filt + 1, p);
			break;
		case 1:	/* expect arg */
			filter_entries[num_filt][sub_filt].str = strdup(p);
			strupr(filter_entries[num_filt][sub_filt].str);
			expect_cmd = 0;
			if (++sub_filt == MAX_SUB)
				do_init_err("Too many fields in filter #%d. (%d max.)\n", num_filt + 1, MAX_SUB);
			break;
		}
	}

	if (!got_args) {
		do_init_err("Syntax error in filter #%d. Actions need (space separated) arguments.\n", num_filt + 1);
	} else if (expect_cmd != 1) {
		if (sub_filt) {
			filter_entries[num_filt][sub_filt].matchfunc = NULL;
			filter_entries[num_filt] = do_realloc(filter_entries[num_filt], (sub_filt + 1) * sizeof(struct filter));
		}
		return (0);
	}
	return (-1);
}

static void save_filter(FILE *fp, char *call)
{
	FILE *savefp;
	char fname[20] = "REQLST/", *p;

	p = stpcpy(&fname[7], call);
	strcpy(p, ".flt");

	if ((savefp = fopen(fname, "wt")) == NULL)
		do_init_err("Can't save filter in %s: err #%d\n", fname, errno);

	fputs("-----\n", savefp);

	while (fgets(line, sizeof(line), fp) != NULL) {
		char *p;

		if ((p = strpbrk(line, "\r\n")))
			*p = '\0';
		if (!strncmp(line, "-----", 5) || !strcmp(line, "/EX"))
			break;
		if (!strcasecmp(line, "SAVE") || !strcasecmp(line, "SHOW") || line[0] == '\0')
			continue;
		fprintf(savefp, "%s\n", line);
	}

	fclose(savefp);
}

static void delete_filter(char *call)
{
	char fname[20] = "REQLST/", *p;

	p = stpcpy(&fname[7], call);
	strcpy(p, ".flt");
	unlink(fname);
}

void restore_saved_filter(char *call)
{
	FILE *fp;
	char fname[20];

	sprintf(fname, saveflt_fmt, call);
	if ((fp = fopen(fname, "rt")) == NULL)
		return;
	init_filter(fp);
	fclose(fp);
}

int init_filter(FILE *fp)
{
	int flt_start = 0;
	long flt_startpos = 0;

	if (setjmp(env))
		return (-1);

	if (filter_entries == NULL)
		filter_entries = do_malloc(MAX_FLT * sizeof(struct filter *));

	while (fgets(line, sizeof(line), fp) != NULL) {
		char *p;

		if (line[0] == '\n')
			continue;
		if ((p = strpbrk(line, "\r\n")))
			*p = '\0';
		if (!strcmp(line, "/EX"))
			break;
		if (!strncmp(line, "-----", 5)) {
			if (!flt_start) {
				flt_start = 1;
				flt_startpos = ftell(fp);
				continue;
			} else
				break;
		} else {
			if (!flt_start)
				continue;
		}
		if (line[0] == '-' || line[0] == '+') {
			if (add_entry(line) == 0)
				if (++num_filt == MAX_FLT)
					do_init_err("Too many filter lines. (%d max.)\n", MAX_FLT);
		} else
			do_cmd(line);
	}

	if (num_filt) {
		filter_entries[num_filt] = NULL;
		filter_entries = do_realloc(filter_entries, (num_filt + 1) * sizeof(struct filter *));
	}

	if (del_filt)
		delete_filter(user_call);
	else if (save_filt && flt_startpos > 0) {
		fseek(fp, flt_startpos, SEEK_SET);
		save_filter(fp, user_call);
	}

	return (num_filt);
}

int test_list_entry(bullist *datap)
{
	int i, j, sub_filt, matched = 0;

	if (!num_filt)
		return (1);

	strcpy(msg.dest, datap->desti);
	strncpy(msg.bbs, datap->bbsv, 6);
	msg.bbs[6] = '\0';
	strcpy(msg.orig, datap->exped);
	strcpy(msg.bid, datap->bid);
	*msg.flags = datap->free[0];
	*msg.loc = datap->bbsf[0];
	for (j = 0; datap->titre[j] && j < sizeof(msg.title); j++)
		msg.title[j] = toupper(datap->titre[j]);
	msg.title[j] = '\0';

	for (i = 0; filter_entries[i] != NULL; i++) {
		matched = 0;
		for (sub_filt = 0; filter_entries[i][sub_filt].matchfunc; sub_filt++) {
			if (matched && filter_entries[i][sub_filt].mode != EXCEPT)
				continue;
			if (filter_entries[i][sub_filt].matchfunc(filter_entries[i][sub_filt].field, filter_entries[i][sub_filt].str)) {
				if (filter_entries[i][sub_filt].mode == EXCEPT) {
					if (matched)
						return (!filter_entries[i][0].mode);
				} else
					matched = 1;
			}
		}
		if (matched)
			break;
	}

	return ((!matched) ? 1 : filter_entries[i][0].mode);
}
